@@ -4,13 +4,19 @@ Base functionality of meshed
 from collections import Counter
 from dataclasses import dataclass, field, fields
 from functools import partial, cached_property
-from typing import Callable, MutableMapping, Iterable, Union, Sized, Sequence
+from typing import Callable, MutableMapping, Iterable, Union, Sized, Sequence, Literal
 
 from i2 import Sig, call_somewhat_forgivingly
-from i2.signatures import ch_variadics_to_non_variadic_kind
+from i2.signatures import (
+    ch_variadics_to_non_variadic_kind,
+    CallableComparator,
+    compare_signatures,
+)
 from meshed.util import ValidationError, NameValidationError, mk_func_name
 from meshed.itools import add_edge
 
+BindInfo = Literal['var_nodes', 'params', 'hybrid']
+
 
 def underscore_func_node_names_maker(func: Callable, name=None, out=None):
     """This name maker will resolve names in the following fashion:
@@ -108,6 +114,8 @@ def handle_variadics(func):
 # TODO: When 3.10, look into and possibly use match_args in to_dict and from_dict
 # TODO: Make FuncNode immutable (is there a way to use frozen=True with post_init?)
 # TODO: How to get a safe hash? Needs to be immutable only?
+# TODO: FuncNode(func_node) gives us FuncNode(scope -> ...). Should we have it be
+#  FuncNode.from_dict(func_node.to_dict()) instead?
 # @dataclass(eq=True, order=True, unsafe_hash=True)
 @dataclass(order=True)
 class FuncNode:
@@ -170,7 +178,7 @@ class FuncNode:
     like this:
 
     >>> scope = {'item_price': 3.5, 'num_of_items': 2}
-    >>> func_node(scope)  # see that it returns 7.0
+    >>> func_node.call_on_scope(scope)  # see that it returns 7.0
     7.0
     >>> scope  # but also wrote this in the scope
     {'item_price': 3.5, 'num_of_items': 2, 'multiply': 7.0}
@@ -208,6 +216,7 @@ class FuncNode:
 
     """
 
+    # TODO: Make everything but func keyword-only (check for non-keyword usage before)
     func: Callable
     name: str = field(default=None)
     bind: dict = field(default_factory=dict)
@@ -240,10 +249,19 @@ class FuncNode:
 
         self.node_validator(self)
 
-    def synopsis_string(self, bind_info='values'):
+    # TODO: BindInfo lists only three unique behaviors, but there are seven actual
+    #  possible values for bind_info. All the rest are convenience aliases. Is this
+    #  a good idea? The hesitation here comes from the fact that the values/keys
+    #  language describes the bind data structure (dict), but the var_nodes/params
+    #  language describes their contextual use. If had to choose, I'd chose the latter.
+    def synopsis_string(self, bind_info: BindInfo = 'values'):
         """
 
-        :param bind_info:
+        :param bind_info: How to represent the bind in the synopsis string. Could be:
+            - 'values', `var_nodes` or `varnodes`: the values of the bind (default).
+            - 'keys' or 'params': the keys of the bind
+            - 'hybrid': the keys of the bind, but with the values that are the same as
+                the keys omitted.
         :return:
 
         >>> fn = FuncNode(
@@ -270,6 +288,8 @@ class FuncNode:
             return f"{','.join(gen())} -> {self.name} " f'-> {self.out}'
         elif bind_info in {'keys', 'params'}:
             return f"{','.join(self.bind.keys())} -> {self.name} " f'-> {self.out}'
+        else:
+            raise ValueError(f'Unknown bind_info: {bind_info}')
 
     def __repr__(self):
         return f'FuncNode({self.synopsis_string(bind_info="hybrid")})'
@@ -291,7 +311,7 @@ class FuncNode:
         return output
 
     def _hash_str(self):
-        """Design ideo.
+        """Design idea.
         Attempt to construct a hash that reflects the actual identity we want.
         Need to transform to int. Only identifier chars alphanumerics and underscore
         and space are used, so could possibly encode as int (for __hash__ method)
@@ -309,7 +329,11 @@ class FuncNode:
 
     def __call__(self, scope):
         """Deprecated: Don't use. Might be a normal function with a signature"""
-        return self.call_on_scope(scope)
+        from warnings import warn
+
+        raise DeprecationWarning(f'Deprecated. Use .call_on_scope(scope) instead.')
+        # warn(f'Deprecated. Use .call_on_scope(scope) instead.', DeprecationWarning)
+        # return self.call_on_scope(scope)
 
     def to_dict(self):
         """The inverse of from_dict: FuncNode.from_dict(fn.to_dict()) == fn"""
@@ -320,6 +344,33 @@ class FuncNode:
         """The inverse of to_dict: Make a ``FuncNode`` from a dictionary of init args"""
         return cls(**dictionary)
 
+    def ch_attrs(self, **new_attrs_values):
+        """Returns a copy of the func node with some of its attributes changed
+
+        >>> def plus(a, b):
+        ...     return a + b
+        ...
+        >>> def minus(a, b):
+        ...     return a - b
+        ...
+        >>> fn = FuncNode(func=plus, out='sum')
+        >>> fn.func == plus
+        True
+        >>> fn.name == 'plus'
+        True
+        >>> new_fn = fn.ch_attrs(func=minus)
+        >>> new_fn.func == minus
+        True
+        >>> new_fn.synopsis_string() == 'a,b -> plus -> sum'
+        True
+        >>>
+        >>>
+        >>> newer_fn = fn.ch_attrs(func=minus, name='sub', out='difference')
+        >>> newer_fn.synopsis_string() == 'a,b -> sub -> difference'
+        True
+        """
+        return ch_func_node_attrs(self, **new_attrs_values)
+
     @classmethod
     def has_as_instance(cls, obj):
         """Verify if ``obj`` is an instance of a FuncNode (or specific sub-class).
@@ -339,7 +390,7 @@ class FuncNode:
 class Mesh:
     func_nodes: Iterable[FuncNode]
 
-    def synopsis_string(self, bind_info='values'):
+    def synopsis_string(self, bind_info: BindInfo = 'values'):
         return '\n'.join(
             func_node.synopsis_string(bind_info) for func_node in self.func_nodes
         )
@@ -406,12 +457,15 @@ def is_func_node(obj) -> bool:
     # A weaker check than an isinstance(obj, FuncNode), which fails when we're
     # developing (therefore changing) FuncNode definition (without relaunching python
     # kernel). This is to be used instead, at least during development times
-    # TODO: Replace with isinstance(obj, FuncNode) is this when development
-    #  stabalizes
+    # TODO: Replace with isinstance(obj, FuncNode) is this when development stabalizes
+    #  See: https://github.com/i2mint/meshed/discussions/57
     # return isinstance(obj, FuncNode)
     cls = type(obj)
     if cls is not type:
-        return any(getattr(x, '__name__', '') == 'FuncNode' for x in cls.mro())
+        try:
+            return any(getattr(x, '__name__', '') == 'FuncNode' for x in cls.mro())
+        except Exception:
+            return isinstance(obj, FuncNode)
     else:
         return False
 
@@ -433,7 +487,7 @@ def get_init_params_of_instance(obj):
 
 
 def ch_func_node_attrs(fn: FuncNode, **new_attrs_values):
-    """Returns a copy of the func node with some of it's attributes changed
+    """Returns a copy of the func node with some of its attributes changed
 
     >>> def plus(a, b):
     ...     return a + b
@@ -467,10 +521,6 @@ def ch_func_node_attrs(fn: FuncNode, **new_attrs_values):
     return FuncNode(**fn_kwargs)
 
 
-def compare_signatures(func1, func2):
-    return Sig(func1) == Sig(func2)
-
-
 def raise_signature_mismatch_error(fn, func):
     raise ValueError(
         'You can only change the func of a FuncNode with a another func if the '
@@ -481,19 +531,65 @@ def raise_signature_mismatch_error(fn, func):
     )
 
 
+# from i2.signatures import keyed_comparator, SignatureComparator
+# if compare_func is None:
+#     compare_func = keyed_comparator(signature_comparator, key=Sig)
+
+
+def _ch_func_node_func(fn: FuncNode, func: Callable):
+    return ch_func_node_attrs(fn, func=func)
+
+
 def ch_func_node_func(
     fn: FuncNode,
     func: Callable,
     *,
-    compare_func=compare_signatures,
+    func_comparator: CallableComparator = compare_signatures,
+    ch_func_node=_ch_func_node_func,
     alternative=raise_signature_mismatch_error,
 ):
-    if compare_func(fn.func, func):
-        return ch_func_node_attrs(fn, func=func)
+    if func_comparator(fn.func, func):
+        return ch_func_node(fn, func=func)
     else:
         return alternative(fn, func)
 
 
+def _new_bind(fnode, new_func):
+    old_sig = Sig(fnode.func)
+    new_sig = Sig(new_func)
+    old_bind: dict = fnode.bind
+    old_to_new_names_map = dict(zip(old_sig.names, new_sig.names))
+    # TODO: assert some health stats on old_to_new_names_map
+    new_bind = {old_to_new_names_map[k]: v for k, v in old_bind.items()}
+    return new_bind
+
+
+# TODO: Add more control (signature comparison, rebinding rules, renaming rules...)
+# TODO: For example, can rebind to a function with different defaults, which are ignored.
+#  Should we allow this? Should we allow to specify how to handle this?
+# TODO: Should we include this in FuncNode as .ch_func(func)?
+#  Possibly with an argument that specifies how to handle details, aligned with the
+#  DAG.ch_funcs method. See ch_func_node_func.
+def rebind_to_func(fnode: FuncNode, new_func: Callable):
+    """Replaces ``fnode.func`` with ``new_func``, changing the ``.bind`` accordingly.
+
+    >>> fn = FuncNode(lambda x, y: x + y, bind={'x': 'X', 'y': 'Y'})
+    >>> fn.call_on_scope(dict(X=2, Y=3))
+    5
+    >>> new_fn = rebind_to_func(fn, lambda a, b, c=0: a * (b + c))
+    >>> new_fn.call_on_scope(dict(X=2, Y=3))
+    6
+    >>> new_fn.call_on_scope(dict(X=2, Y=3, c=1))
+    8
+    """
+    new_bind = _new_bind(fnode, new_func)
+    return fnode.ch_attrs(func=new_func, bind=new_bind)
+
+
+def insert_func_if_compatible(func_comparator: CallableComparator = compare_signatures):
+    return partial(ch_func_node_func, func_comparator=func_comparator)
+
+
 def _keys_and_values_are_strings_validation(d: dict):
     for k, v in d.items():
         if not isinstance(k, str):
