@@ -136,7 +136,7 @@ from collections import defaultdict
 
 from dataclasses import dataclass, field
 from itertools import chain
-from operator import attrgetter
+from operator import attrgetter, eq
 from typing import (
     Callable,
     MutableMapping,
@@ -159,9 +159,12 @@ from i2.signatures import (
     empty,
     Sig,
     sort_params,
+    # SignatureComparator,
+    CallableComparator,
 )
 from meshed.base import (
     FuncNode,
+    BindInfo,
     ch_func_node_func,
     validate_that_func_node_names_are_sane,
     _mk_func_nodes,
@@ -169,6 +172,7 @@ from meshed.base import (
     is_func_node,
     FuncNodeAble,
     func_node_transformer,
+    compare_signatures,
 )
 
 from meshed.util import (
@@ -401,6 +405,7 @@ def modified_func_node(func_node, **modifications) -> FuncNode:
 
 from i2 import partialx
 
+
 # TODO: doctests
 def partialized_funcnodes(func_nodes, **keyword_defaults):
     for func_node in func_nodes:
@@ -560,7 +565,6 @@ class DAG:
         return cls(func_nodes)
 
     def bindings_cleaner(self):
-
         self.func_nodes = mk_nodes_names_unique(self.func_nodes)
         funcnodes_names = [node.name for node in self.func_nodes]
         func = lambda v: self._func_node_for[v].out
@@ -592,6 +596,25 @@ class DAG:
         # extract the desired final output and return it
         return self.extract_output_from_scope(scope, self.leafs)
 
+    def _preprocess_scope(self, scope):
+        """Take care of the stuff that needs to be taking care of before looping
+        though the func_nodes and calling them on scope. Namely:
+
+        - If scope is None, create a new one calling self.new_scope()
+        - If self.cache_last_scope is True, remember the scope in self.last_scope
+
+        """
+        if scope is None:
+            scope = self.new_scope()  # fresh new scope
+        if self.cache_last_scope:
+            self.last_scope = scope  # just to remember it, for debugging purposes ONLY!
+        return scope
+
+    def _call_func_nodes_on_scope(self, scope):
+        """Loop over ``func_nodes`` calling func_node.call_on_scope on scope."""
+        for func_node in self.func_nodes:
+            func_node.call_on_scope(scope)
+
     def call_on_scope(self, scope=None):
         """Calls the func_nodes using scope (a dict or MutableMapping) both to
         source it's arguments and write it's results.
@@ -602,13 +625,8 @@ class DAG:
         log read and/or writes to specific keys, or disallow overwriting to an existing
         key (useful for pipeline sanity), etc.
         """
-        if scope is None:
-            scope = self.new_scope()  # fresh new scope
-        if self.cache_last_scope:
-            self.last_scope = scope  # just to remember it, for debugging purposes ONLY!
-
-        for func_node in self.func_nodes:
-            func_node.call_on_scope(scope)
+        scope = self._preprocess_scope(scope)
+        self._call_func_nodes_on_scope(scope)
 
     # def clone(self, *args, **kwargs):
     #     """Use args, kwargs to make an instance, using self attributes for
@@ -954,7 +972,17 @@ class DAG:
         """
         yield from self.func_nodes
 
-    def ch_funcs(self, **func_mapping: Callable):
+    # Note: signature_comparator is position only to not conflict with any of the
+    #  func_mapping keys.
+    def ch_funcs(
+        self,
+        # func_comparator: CallableComparator = compare_signatures,
+        ch_func_node_func: Callable[
+            [FuncNode, Callable, CallableComparator], FuncNode
+        ] = ch_func_node_func,
+        /,
+        **func_mapping: Callable,
+    ) -> 'DAG':
         """
         Change some of the functions in the DAG.
         More preciseluy get a copy of the DAG where in some of the functions have
@@ -989,8 +1017,97 @@ class DAG:
 
         You can reference the ``FuncNode`` you want to change through its ``.name`` or
         ``.out`` attribute (both are unique to this ``FuncNode`` in a ``DAG``).
+
+        >>> from i2 import Sig
+        >>>
+        >>> dag = DAG([
+        ...     FuncNode(lambda a, b: a + b, name='f'),
+        ...     FuncNode(lambda y=1, z=2: y * z, name='g', bind={'z': 'f'})
+        ... ])
+        >>>
+        >>> Sig(dag)
+        <Sig (a, b, f=2, y=1)>
+        >>>
+
+        If you replace by a different function with exactly the same signature,
+        all goes well:
+
+        >>> dag.ch_funcs(g=lambda y=1, z=2: y / z)
+        DAG(func_nodes=[FuncNode(a,b -> f -> _f), FuncNode(z=_f,y -> g -> _g)], name=None)
+
+        But if you change the signature, even slightly you get an error.
+
+        Here we didn't include the defaults:
+
+        >>> dag.ch_funcs(g=lambda y, z: y / z)  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
+        Traceback (most recent call last):
+          ...
+        ValueError: You can only change the func of a FuncNode with a another func if the signatures match.
+          ...
+
+        Here we include defaults, but ``z``'s is different:
+
+        >>> dag.ch_funcs(g=lambda y=1, z=200: y / z)  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
+        Traceback (most recent call last):
+          ...
+        ValueError: You can only change the func of a FuncNode with a another func if the signatures match.
+          ...
+
+        Here the defaults are exactly the same, but the order of parameters is
+        different:
+
+        >>> dag.ch_funcs(g=lambda z=2, y=1: y / z)  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
+        Traceback (most recent call last):
+          ...
+        ValueError: You can only change the func of a FuncNode with a another func if the signatures match.
+          ...
+
+        This validation of the functions controlled by the ``func_comparator``
+        argument. By default this is the ``compare_signatures`` which compares the
+        signatures of the functions in the strictest way possible.
+        The is the right choice for a default since it will get you out of trouble
+        down the line.
+
+        But it's also annoying in many situations, and in those cases you should
+        specify the ``func_comparator`` that makes sense for your context.
+
+        Since most of the time, you'll want to compare functions solely based on
+        their signature, we provide a ``compare_signatures`` allows you to control the
+        signature comparison through a ``signature_comparator`` argument.
+
+        >>> from meshed import compare_signatures
+        >>> from functools import partial
+        >>> on_names = lambda sig1, sig2: list(sig1.parameters) == list(sig2.parameters)
+        >>> same_names = partial(compare_signatures, signature_comparator=on_names)
+        >>> ch_fnode = partial(ch_func_node_func, func_comparator=same_names)
+        >>> d = dag.ch_funcs(ch_fnode, g=lambda y, z: y / z);
+        >>> Sig(d)
+        <Sig (a, b, y)>
+        >>> d(2, 3, 4)
+        0.8
+
+        And this one works too:
+
+        >>> d = dag.ch_funcs(ch_fnode, g=lambda y=1, z=200: y / z);
+
+        But our ``same_names`` function compared names including their order.
+        If we want a function with the signature ``(z=2, y=1)`` to be able to be
+        "injected" we'll need a different comparator:
+
+        >>> _names = lambda sig1, sig2: set(sig1.parameters) == set(sig2.parameters)
+        >>> same_set_of_names = partial(
+        ...     compare_signatures,
+        ...     signature_comparator=(
+        ...         lambda sig1, sig2: set(sig1.parameters) == set(sig2.parameters)
+        ...     )
+        ... )
+        >>> ch_fnode2 = partial(ch_func_node_func, func_comparator=same_set_of_names)
+        >>> d = dag.ch_funcs(ch_fnode2, g=lambda z=2, y=1: y / z);
+
         """
-        return ch_funcs(self, func_mapping=func_mapping,)
+        return ch_funcs(
+            self, func_mapping=func_mapping, ch_func_node_func=ch_func_node_func
+        )
 
         # _validate_func_mapping(func_mapping, self)
         #
@@ -1065,7 +1182,7 @@ class DAG:
                 # is determined by what ``parameter_merge`` specified, which is,
                 # by default, strict (everything needs to be the same, or
                 # ``parameter_merge`` with raise an error.)
-                yield self.parameter_merge(params_with_name_changed_to_src_name)
+                yield self.parameter_merge(*params_with_name_changed_to_src_name)
 
     # TODO: Find more representative (and possibly shorter) doctest:
     @property
@@ -1094,6 +1211,35 @@ class DAG:
             for k, v in self.graph.items()
         }
 
+    def _prepare_other_for_addition(self, other):
+        if isinstance(other, DAG):
+            other = list(other.func_nodes)
+        elif isinstance(other, int) and other == 0:
+            # Note: This is so that we can use sum(dags) to get a union of dags without
+            # having to specify the initial DAG() value of sum (which is 0 by default).
+            other = DAG()
+        else:
+            other = list(DAG(other).func_nodes)
+
+        return other
+
+    def __radd__(self, other):
+        """A union of DAGs. See ``__add__`` for more details.
+
+        >>> dag = sum([DAG(list), DAG(tuple)])
+        >>> print(dag.synopsis_string(bind_info='hybrid'))
+        iterable -> list_ -> list
+        iterable -> tuple_ -> tuple
+        >>> dag([1,2,3])
+        ([1, 2, 3], (1, 2, 3))
+
+        """
+        # We could have just returned self + other to be commutative, but perhaps
+        # we would like to control some orders of things via the order of addition
+        # (thinkg list addition versus set addition for example), so instead we write
+        # the explicit code:
+        return DAG(self._prepare_other_for_addition(other) + list(self.func_nodes))
+
     def __add__(self, other):
         """A union of DAGs.
 
@@ -1106,11 +1252,7 @@ class DAG:
         >>> dag([1,2,3])
         ([1, 2, 3], (1, 2, 3))
         """
-        if isinstance(other, DAG):
-            other = list(other.func_nodes)
-        else:
-            other = list(DAG(other).func_nodes)
-        return DAG(list(self.func_nodes) + other)
+        return DAG(list(self.func_nodes) + self._prepare_other_for_addition(other))
 
     def copy(self, renamer=numbered_suffix_renamer):
         return DAG(ch_names(self.func_nodes, renamer=renamer))
@@ -1312,6 +1454,7 @@ class DAG:
             scope={'a': 1, 'b': 2, 'c': 3, 'd': 4, 'f': 3, 'g': 12, 'h': 9}
 
         """
+
         # TODO: Add feedback callable validation
         @Sig(self)
         def launch_debugger(*args, **kwargs):
@@ -1327,7 +1470,7 @@ class DAG:
 
     # ------------ display --------------------------------------------------------------
 
-    def synopsis_string(self, bind_info='values'):
+    def synopsis_string(self, bind_info: BindInfo = 'var_nodes'):
         return '\n'.join(
             func_node.synopsis_string(bind_info) for func_node in self.func_nodes
         )
@@ -1395,6 +1538,9 @@ class DAG:
     def sig(self, value):
         self.__signature__ = value
 
+    def find_funcs(self, filt: Callable[[FuncNode], bool] = None) -> Iterable[Callable]:
+        return (func_node.func for func_node in filter(filt, self.func_nodes))
+
 
 # These are the defaults used in lined.
 # TODO: Merge some of the functionalities around graph displays in lined and meshed
@@ -1549,7 +1695,6 @@ def dot_lines_of_func_nodes(
 
 
 def dot_lines_of_func_node(func_node: FuncNode, **kwargs):
-
     out = func_node.out
 
     func_id = func_node.name
@@ -1574,7 +1719,7 @@ def _add_new_line_if_none(s: str):
     return s
 
 
-def print_dag_string(dag: DAG, bind_info: str = 'hybrid'):
+def print_dag_string(dag: DAG, bind_info: BindInfo = 'hybrid'):
     print(dag.synopsis_string(bind_info=bind_info))
 
 
@@ -1620,6 +1765,7 @@ names_and_outs = partial(attribute_vals, attrs=('name', 'out'), egress=chain)
 
 DagAble = Union[DAG, Iterable[FuncNodeAble]]
 
+
 # TODO: Extract hardcoded ".name or .out" condition so indexing/condition can be
 #  controlled by user.
 def _validate_func_mapping(func_mapping: FuncMapping, func_nodes: DagAble):
@@ -1680,6 +1826,8 @@ def _validate_func_mapping(func_mapping: FuncMapping, func_nodes: DagAble):
 
 FuncMappingValidator = Callable[[FuncMapping, DagAble], None]
 
+
+# TODO: Redesign. Is terrible both in interface and code.
 # TODO: Merge with DAG, or with Mesh (when it exists)
 # TODO: Make it work with any FuncNode Iterable
 # TODO: extract egress functionality to decorator?
@@ -1688,8 +1836,14 @@ def ch_funcs(
     func_nodes: DagAble = None,
     *,
     func_mapping: FuncMapping = (),
-    ch_func_node_func: Callable[[FuncNode, Callable], FuncNode] = ch_func_node_func,
     validate_func_mapping: Optional[FuncMappingValidator] = _validate_func_mapping,
+    # TODO: Design. Don't like the fact that ch_func_node_func needs a slot for
+    #  func_comparator, which is then given later. Perhaps only ch_func_node_func should
+    #  should be given (and it contains the func_comparator)
+    ch_func_node_func: Callable[
+        [FuncNode, Callable, CallableComparator], FuncNode
+    ] = ch_func_node_func,
+    # func_comparator: CallableComparator = compare_signatures,
 ):
     """Function (and decorator) to change the functions of func_nodes according to
     the specification of a func_mapping whose keys are ``.name`` or ``.out`` values
@@ -1717,11 +1871,11 @@ def ch_funcs(
 
     # TODO: Optimize (for example, use self._func_node_for)
     def ch_func(dag, key, func):
+        condition = lambda fn: fn.name == key or fn.out == key  # TODO: interface ctrl?
+        replacement = lambda fn: ch_func_node_func(fn, func,)
         return DAG(
             replace_item_in_iterable(
-                dag.func_nodes,
-                condition=lambda fn: fn.name == key or fn.out == key,
-                replacement=lambda fn: ch_func_node_func(fn, func),
+                dag.func_nodes, condition=condition, replacement=replacement,
             )
         )
 
