@@ -487,6 +487,8 @@ def _extract_name_from_single_func_def(src: str, default=None):
 FuncSource = Union[Callable[[str], Callable], Mapping[str, Callable]]
 
 
+# TODO: Make code_to_fnodes more flexible (not need to be enclosed in a function
+#  definition)
 @double_up_as_factory
 def code_to_fnodes(
     src=None,
@@ -522,6 +524,62 @@ def code_to_digraph(src):
 simple_code_to_digraph = code_to_digraph  # back-compatability alias
 
 
+# import re
+# from typing import Tuple, Iterable, Iterator
+# from meshed import FuncNode, code_to_dag, code_to_fnodes, DAG
+
+extract_tokens = re.compile('\w+').findall
+
+
+def triples_to_fnodes(triples: Iterable[Tuple[str, str, str]]) -> Iterable[FuncNode]:
+    """Converts an iterable of func call triples to an iterable of ``FuncNode``s.
+    (Which in turn can be converted to a ``DAG``.)
+
+    Note how the python identifiers are extracted (on the basis of "an unbroken
+    sequence of alphanumerical (and underscore) characters", ignoring all other
+    characters).
+
+    >>> from meshed import DAG
+    >>> dag = DAG(
+    ...     triples_to_fnodes(
+    ...     [
+    ...         ('alpha bravo', 'charlie', 'delta echo'),
+    ...         (' foxtrot  &^$#', 'golf', '  alpha,  echo'),
+    ...     ])
+    ... )
+    >>> print(dag.synopsis_string())
+    delta,echo -> charlie -> alpha__bravo
+    alpha__bravo -> alpha__0 -> alpha
+    alpha__bravo -> bravo__1 -> bravo
+    alpha,echo -> golf -> foxtrot
+    """
+    code = '\n\t'.join(_triple_to_func_call_str(*triple) for triple in triples)
+    code = f'def main():\n\t{code}'
+    return code_to_fnodes(code)
+
+
+def _triple_to_func_call_str(
+    outputs: Union[str, Iterable[str]],
+    func_name: str,
+    inputs: Union[str, Iterable[str]],
+) -> str:
+    """Converts a `(outputs, func_name, inputs)` triple to a function call string.
+
+    >>> _triple_to_func_call_str(('a', 'b'), 'func', ('c', 'd'))
+    'a, b = func(c, d)'
+    """
+    outputs = _ensure_tokens_iterable(outputs)
+    inputs = _ensure_tokens_iterable(inputs)
+    return f"{', '.join(outputs)} = {func_name}({', '.join(inputs)})"
+
+
+def _ensure_tokens_iterable(tokens):
+    if isinstance(tokens, str):
+        return extract_tokens(tokens)
+    else:
+        return tokens
+
+
 def _ensure_func_src(
     func_src: FuncSource, use_place_holder_fallback=False
 ) -> Callable[[str], Callable]:
@@ -574,3 +632,70 @@ def lined_dag(funcs):
         pairs = zip(names, names[1:])
         dag = dag.add_edges(pairs)
         return dag
+
+
+from typing import Callable, Mapping, Iterable
+
+NamedFuncs = Mapping[str, Callable]
+
+
+def named_funcs_to_func_nodes(named_funcs: NamedFuncs) -> Iterable[FuncNode]:
+    """Make ``FuncNode``s from keyword arguments, using the key as the ``.out`` of the
+    ``FuncNode`` and the value as the ``.func`` of the ``FuncNode``.
+
+    Example use: To get from ``Slabs`` to ``DAG``.
+
+    >>> from meshed import DAG
+    >>> func_nodes = list(named_funcs_to_func_nodes(dict(
+    ...     a=lambda x: x + 1,
+    ...     b=lambda a: a + 2,
+    ...     c=lambda a, b: a * b)
+    ... ))
+    >>> dag = DAG(func_nodes)
+    >>> dag(x=3)
+    24
+
+    The inverse of this function is ``func_nodes_to_named_funcs``.
+
+    >>> named_funcs = func_nodes_to_named_funcs(dag.func_nodes)
+    >>> dag2 = DAG(named_funcs_to_func_nodes(named_funcs))
+    >>> assert dag2(x=3) == dag(x=3) == 24
+
+    """
+    return (
+        FuncNode(func, name=f'{out}_', out=out) for out, func in named_funcs.items()
+    )
+
+
+def func_nodes_to_named_funcs(func_nodes: Iterable[FuncNode]) -> NamedFuncs:
+    """Make some components (kwargs) based on the ``.out`` and ``.func`` of the
+    ``FuncNode``s.
+
+    Example use: To get from ``DAG`` to ``Slabs``.
+
+    >>> from meshed import DAG, FuncNode
+    >>> dag = DAG([
+    ...     FuncNode(lambda x: x + 1, out='a'),
+    ...     FuncNode(lambda a: a + 2, out='b',),
+    ...     FuncNode(lambda a, b: a * b, out='c'),
+    ... ])
+    >>> dag(x=10)
+    143
+    >>> named_funcs = func_nodes_to_named_funcs(dag.func_nodes)
+    >>> isinstance(named_funcs, dict)
+    True
+    >>> list(named_funcs)
+    ['a', 'b', 'c']
+    >>> callable(named_funcs['a'])
+    True
+    >>> assert dag.find_func_node('a').func(3) == named_funcs['a'](3) ==4
+
+    The inverse of this function is ``named_funcs_to_func_nodes``.
+
+    >>> func_nodes = list(named_funcs_to_func_nodes(named_funcs))
+    >>> dag2 = DAG(func_nodes)
+    >>> assert dag2(x=3) == dag(x=3) == 24
+
+
+    """
+    return {node.out: node.func for node in func_nodes}
