@@ -4,12 +4,56 @@ from functools import partial, wraps
 from inspect import Parameter
 from typing import Callable, Any, Union, Iterator, Optional, Iterable, Mapping, TypeVar
 
-from i2 import Sig, name_of_obj, Literal, FuncFanout, Pipe
+from i2 import Sig, name_of_obj, LiteralVal, FuncFanout, Pipe
 from operator import itemgetter
 
 T = TypeVar('T')
 
 
+def provides(*var_names: str) -> Callable[[Callable], Callable]:
+    """Decorator to assign ``var_names`` to a ``provides_`` attribute of function.
+    
+    This is meant to be used to indicate to a mesh what var nodes a function can source
+    values for.
+
+    >>> @provides('a', 'b')
+    ... def f(x):
+    ...     return x + 1
+    >>> f._provides
+    ('a', 'b')
+
+    If no ``var_names`` are given, then the function name is used as the var name:
+
+    >>> @provides()
+    ... def g(x):
+    ...     return x + 1
+    >>> g._provides
+    ('g',)
+
+    If ``var_names`` contains ``'_'``, then the function name is used as the var name
+    for that position:
+    
+    >>> @provides('b', '_')
+    ... def h(x):
+    ...     return x + 1
+    >>> h._provides
+    ('b', 'h')
+
+    """
+
+    def add_provides_attribute(func):
+        if not var_names:
+            var_names_ = (name_of_obj(func),)
+        else:
+            var_names_ = tuple(
+                [x if x != '_' else name_of_obj(func) for x in var_names]
+            )
+        func._provides = var_names_
+        return func
+
+    return add_provides_attribute
+
+
 def if_then_else(if_func, then_func, else_func, *args, **kwargs):
     """
     Tool to "functionalize" the if-then-else logic.
@@ -366,7 +410,8 @@ func_name: FunctionNamer
 
 def func_name(func) -> str:
     """The func.__name__ of a callable func, or makes and returns one if that fails.
-    To make one, it calls unamed_func_name which produces incremental names to reduce the chances of clashing"""
+    To make one, it calls unamed_func_name which produces incremental names to reduce the chances of clashing
+    """
     try:
         name = func.__name__
         if name == '<lambda>':
@@ -705,6 +750,7 @@ def _not_callable(obj):
     return not callable(obj)
 
 
+# Pattern: routing
 # TODO: Replace
 def conditional_trans(
     obj: T, condition: Callable[[T], bool], trans: Callable[[T], Any]
@@ -724,13 +770,13 @@ def conditional_trans(
     what, wrap the input with ``Literal``
 
     >>> # from meshed import Literal
-    >>> conditional_trans(Literal('10'), str.isnumeric, float)
+    >>> conditional_trans(LiteralVal('10'), str.isnumeric, float)
     '10'
 
     """
     # TODO: Maybe make Literal checking less sensitive to isinstance checks, using
     #   hasattr instead for example.
-    if isinstance(obj, Literal):  # If val is a Literal, return its value as is
+    if isinstance(obj, LiteralVal):  # If val is a Literal, return its value as is
         return obj.val
     elif condition(obj):  # If obj satisfies condition, return the alternative_obj
         return trans(obj)
@@ -858,12 +904,17 @@ def extract_items(d: dict, keys: Iterable):
 
 
 ParameterMerger = Callable[[Iterable[Parameter]], Parameter]
-
-conservative_parameter_merge: ParameterMerger
+parameter_merger: ParameterMerger
 
 
-def conservative_parameter_merge(
-    params, same_kind=True, same_default=True, same_annotation=True
+# TODO: Be aware of i2.signatures.param_comparator in
+#  https://github.com/i2mint/i2/blob/2bd43b350a3ae29f1e6c587dbe15d6f536635173/i2/signatures.py#L4247
+#  and related funnctions, which are meant to be a more general approach. Consider
+#  merging parameter_merger to use that general tooling.
+# TODO: Make the ValidationError be even more specific, indicating what parameters
+#  are different and how.
+def parameter_merger(
+    *params, same_name=True, same_kind=True, same_default=True, same_annotation=True
 ):
     """Validates that all the params are exactly the same, returning the first if so.
 
@@ -875,23 +926,41 @@ def conservative_parameter_merge(
 
     But if they're not the same, we need to provide control on which to ignore.
 
+    >>> from inspect import Parameter as P
+    >>> PK = P.POSITIONAL_OR_KEYWORD
+    >>> KO = P.KEYWORD_ONLY
+    >>> parameter_merger(P('a', PK), P('a', PK))
+    <Parameter "a">
+    >>> parameter_merger(P('a', PK), P('different_name', PK), same_name=False)
+    <Parameter "a">
+    >>> parameter_merger(P('a', PK), P('a', KO), same_kind=False)
+    <Parameter "a">
+    >>> parameter_merger(P('a', PK), P('a', PK,  default=42), same_default=False)
+    <Parameter "a">
+    >>> parameter_merger(P('a', PK, default=42), P('a', PK), same_default=False)
+    <Parameter "a=42">
+    >>> parameter_merger(P('a', PK, annotation=int), P('a', PK), same_annotation=False)
+    <Parameter "a: int">
     """
     suggestion_on_error = '''To resolve this you have several choices:
 
     - Change the properties of the param (kind, default, annotation) to be those you 
-      want. For example, you can use ``i2.Sig.ch_param_attrs`` 
-      (or ``i2.Sig.ch_defaults``, ``i2.Sig.ch_kinds``, ``i2.Sig.ch_annotations``)
+      want. For example, you can use ``i2.Sig.ch_param_attrs`` on the signatures 
+      (or ``i2.Sig.ch_names``, ``i2.Sig.ch_defaults``, ``i2.Sig.ch_kinds``, 
+      ``i2.Sig.ch_annotations``)
       to get a function decorator that will do that for you.
     - If you're making a DAG, consider specifying a different ``parameter_merge``.
       For example you can use ``functools.partial`` on 
-      ``i2.dag.conservative_parameter_merge``, fixing ``same_kind``, ``same_default``, 
+      ``meshed.parameter_merger``, fixing ``same_kind``, ``same_default``, 
       and/or ``same_annotation`` to ``False`` to get a more lenient version of it.
+      (See also i2.signatures.param_comparator.)
 
-    See https://github.com/i2mint/meshed/issues/7 (description and comments) for more
+    See https://github.com/i2mint/i2/discussions/63 and 
+    https://github.com/i2mint/meshed/issues/7 (description and comments) for more
     info.
     '''
     first_param, *_ = params
-    if not all(p.name == first_param.name for p in params):
+    if same_name and not all(p.name == first_param.name for p in params):
         raise ValidationError(
             f"Some params didn't have the same name: {params}\n{suggestion_on_error}"
         )
@@ -911,3 +980,8 @@ def conservative_parameter_merge(
             f'{params}\n{suggestion_on_error}'
         )
     return first_param
+
+
+conservative_parameter_merge: ParameterMerger = partial(
+    parameter_merger, same_kind=True, same_default=True, same_annotation=True
+)
